/**
 * Juliac
 * Copyright (C) 2009-2013 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Christophe Munilla
 */
package org.objectweb.fractal.juliac.osgi.revision;

import java.io.InputStream;
import java.lang.reflect.Field;
import java.util.Dictionary;
import java.util.logging.Level;

import org.objectweb.fractal.juliac.osgi.revision.AbstractBundle;
import org.objectweb.fractal.juliac.osgi.revision.BundleContextRevision;
import org.objectweb.fractal.juliac.osgi.revision.BundleRevision;
import org.objectweb.fractal.juliac.osgi.revision.FrameworkListenerRevision;
import org.objectweb.fractal.juliac.osgi.revision.OSGiRevisionException;
import org.objectweb.fractal.juliac.osgi.revision.ServiceListenerRevision;
import org.objectweb.fractal.juliac.osgi.revision.api.BundleContextClassLoader;
import org.objectweb.fractal.juliac.osgi.revision.api.BundleContextRevisionItf;
import org.objectweb.fractal.juliac.osgi.revision.api.BundleRevisionItf;
import org.objectweb.fractal.juliac.osgi.revision.event.BundleListenerRevisionItf;
import org.objectweb.fractal.juliac.osgi.revision.event.FrameworkListenerRevisionItf;
import org.objectweb.fractal.juliac.osgi.revision.event.ServiceListenerRevisionItf;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleEvent;
import org.osgi.framework.BundleListener;
import org.osgi.framework.FrameworkListener;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.ServiceListener;
import org.osgi.framework.ServiceRegistration;

/**
 * BundleContextRevisionItf implementation for Concierge-R3 Framework 
 */
public class ConciergeBundleContext extends BundleContextRevision
{
    
    private ClassLoader bundleClassLoaderParent;
    private final boolean FORCE_BUNDLECLASSLOADER_PARENT;
    
    /**
     * Constructor
     * 
     * @param conciergeRevision
     * @param systemBundleContext
     * @param classLoader
     */
    public ConciergeBundleContext(ConciergeRevision conciergeRevision,
            BundleContext systemBundleContext,
            BundleContextClassLoader classLoader)
    {
        super(conciergeRevision, systemBundleContext, classLoader);
        this.bundleClassLoaderParent = ((ClassLoader)classLoader).getParent();
        FORCE_BUNDLECLASSLOADER_PARENT = (bundleClassLoaderParent != ClassLoader.getSystemClassLoader());        
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.objectweb.fractal.juliac.osgi.revision.api.BundleContextRevisionItf#getBundles()
     */
    @Override
    public BundleRevisionItf<Bundle, BundleContext>[] getBundles()
    {
        BundleRevisionItf<Bundle, BundleContext>[] bundles = super.getBundles();
        
        BundleRevisionItf<Bundle, BundleContext>[] returnedBundles = 
            new BundleRevisionItf[bundles.length - 1];
        
        if(bundles != null)
        {               
            int index = 0;
        
            for(BundleRevisionItf<Bundle,BundleContext> bundle : bundles)
            {
                //if it is the SystemBundle
                if(0 == bundle.getBundleId())
                {
                    //continue
                    continue;
                }
                try
                {
                    registerBundleClassLoaderParent(bundle);
                    returnedBundles[index++] = bundle;
                    
                } catch (IllegalArgumentException e)
                {
                    logger.log(Level.CONFIG,e.getMessage(),e); 
                    
                } catch (SecurityException e)
                {
                    logger.log(Level.CONFIG,e.getMessage(),e); 
                    
                } catch (IllegalAccessException e)
                {
                    logger.log(Level.CONFIG,e.getMessage(),e); 
                    
                } catch (NoSuchFieldException e)
                {
                    logger.log(Level.CONFIG,e.getMessage(),e); 
                }
            }
        }
        return returnedBundles;
    }
    
    
    /**
     * {@inheritDoc}
     * 
     * @see org.objectweb.fractal.juliac.osgi.revision.api.BundleContextRevisionItf#
     * installBundle(java.lang.String)
     */
    @Override
    public BundleRevisionItf<Bundle,BundleContext> 
    installBundle(String location) throws OSGiRevisionException
    {
        BundleRevisionItf<Bundle,BundleContext> bundle = super.installBundle(location);
        try
        {
            registerBundleClassLoaderParent(bundle);
            
        } catch (IllegalArgumentException e)
        {
            logger.log(Level.CONFIG,e.getMessage(),e); 
        } catch (SecurityException e)
        {
            logger.log(Level.CONFIG,e.getMessage(),e); 
        } catch (IllegalAccessException e)
        {
            logger.log(Level.CONFIG,e.getMessage(),e); 
        } catch (NoSuchFieldException e)
        {
            logger.log(Level.CONFIG,e.getMessage(),e); 
        }
        return bundle;
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.objectweb.fractal.juliac.osgi.revision.api.BundleContextRevisionItf#
     * installBundle(java.lang.String, java.io.InputStream)
     */
    @Override
    public BundleRevisionItf<Bundle,BundleContext>
    installBundle(String location, InputStream bundleStream)
            throws OSGiRevisionException
    {
        BundleRevisionItf<Bundle,BundleContext> bundle = super.installBundle(
                location,bundleStream);
        try
        {
            registerBundleClassLoaderParent(bundle);
            
        } catch (IllegalArgumentException e)
        {
            logger.log(Level.CONFIG,e.getMessage(),e); 
        } catch (SecurityException e)
        {
            logger.log(Level.CONFIG,e.getMessage(),e); 
        } catch (IllegalAccessException e)
        {
            logger.log(Level.CONFIG,e.getMessage(),e); 
        } catch (NoSuchFieldException e)
        {
            logger.log(Level.CONFIG,e.getMessage(),e); 
        }
        return bundle;
    }
    
    /**
     * {@inheritDoc}
     * 
     * @see org.objectweb.fractal.juliac.osgi.revision.api.BundleContextRevisionItf#
     * getBundle(int)
     */
    @Override
    public BundleRevisionItf<Bundle,BundleContext> getBundle(long bundleId)
    {
        BundleRevisionItf<Bundle,BundleContext> bundle = super.getBundle(bundleId);
        try
        {
            registerBundleClassLoaderParent(bundle);
            
        } catch (IllegalArgumentException e)
        {
            logger.log(Level.CONFIG,e.getMessage(),e); 
        } catch (SecurityException e)
        {
            logger.log(Level.CONFIG,e.getMessage(),e); 
        } catch (IllegalAccessException e)
        {
            logger.log(Level.CONFIG,e.getMessage(),e); 
        } catch (NoSuchFieldException e)
        {
            logger.log(Level.CONFIG,e.getMessage(),e); 
        }
        return bundle;
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.objectweb.fractal.juliac.osgi.revision.api.BundleContextRevisionItf#
     * getServiceBundle(java.lang.String, java.lang.String)
     */
    @Override
    public BundleRevisionItf<Bundle, BundleContext> getServiceBundle(
            String className, String filter)
    {
        BundleRevisionItf<Bundle,BundleContext> bundle = super.getServiceBundle(className,filter);
        try
        {
            registerBundleClassLoaderParent(bundle);
            
        } catch (IllegalArgumentException e)
        {
            logger.log(Level.CONFIG,e.getMessage(),e); 
            
        } catch (SecurityException e)
        {
            logger.log(Level.CONFIG,e.getMessage(),e); 
            
        } catch (IllegalAccessException e)
        {
            logger.log(Level.CONFIG,e.getMessage(),e); 
            
        } catch (NoSuchFieldException e)
        {
            logger.log(Level.CONFIG,e.getMessage(),e); 
        }
        return bundle;
    }

    /**
     * The Concierge framework's BundleClassLoader class has no parent (it's a 
     * final field but never defined, so we can do it here) and so refers to
     * the running system (FraSCAti fo example) classloader when a search class is 
     * not defined in the Bundle. Here we force the parent ClassLoader to be able to search 
     * classes where they are
     */
    private void registerBundleClassLoaderParent(BundleRevisionItf<Bundle,BundleContext> bundle)
    throws IllegalArgumentException, IllegalAccessException, SecurityException, NoSuchFieldException
    {
        if(bundle == null || !FORCE_BUNDLECLASSLOADER_PARENT)
        {
            return;
        }
        Class<?> bundleClass = bundle.getBundle().getClass();
        Field classLoaderField = bundleClass.getDeclaredField("classloader");
        classLoaderField.setAccessible(true);
        
        Object classloader = classLoaderField.get(bundle.getBundle());
       
        Field parent = ClassLoader.class.getDeclaredField("parent");
        parent.setAccessible(true);
        parent.set(classloader,bundleClassLoaderParent);
    }
    
    /**
     * In the Concierge Framework, if a BundleContext is not associated to a Bundle then 
     * the service registration failed. So, if no Bundle instance is associated 
     * to the current BundleContext one, as no BundleContext is associated to the 
     * SystemBundle object, we use the BundleContext of the frascati-osgi-api Bundle to
     * register the service
     *  
     * @see org.objectweb.fractal.juliac.osgi.revision.api.BundleContextRevisionItf#
     * registerService(java.lang.String, java.lang.Object, java.util.Dictionary)
     */
    @Override
    public int registerService(String className, Object service,
            Dictionary<String, String> properties)
    {
        int registrationId = -1;
        
        if(super.bundleContext.getBundle() == null)
        {
            BundleRevisionItf<Bundle,BundleContext> apiBundle = getBundle(
                    "bundle-frascati-api");
            
            if(apiBundle != null)
            {
                BundleContextRevisionItf<Bundle,BundleContext> bundleContextR = 
                        apiBundle.getBundleContext(); 
                
                registrationId = bundleContextR.registerService(className, 
                        service, properties);
                
            }
        } else
        {
         ServiceRegistration registration = super.bundleContext.registerService(
                className, service, properties);
        
         logger.log(Level.CONFIG,"Register '" + className + "' service : " + registration); 
        
         registrationId = registration.hashCode();
         registrations.put(new Integer(registrationId), registration);
        }
        return registrationId;
    }
    
    /**
     * {@inheritDoc}
     * 
     * cf. registerService method comment
     * 
     * @see org.objectweb.fractal.juliac.osgi.revision.api.BundleContextRevisionItf#
     * registerBundleListener(org.objectweb.fractal.juliac.osgi.revision.event.BundleListenerRevisionItf)
     */
    @Override
    public void registerBundleListener(BundleListenerRevisionItf<Bundle,BundleContext> bundleListener)
    { 
        if(super.bundleContext.getBundle() == null)
        {
            BundleRevisionItf<Bundle,BundleContext> apiBundle = getBundle("bundle-frascati-api");
            if(apiBundle != null)
            {
                BundleContextRevisionItf<Bundle,BundleContext> bundleContextR = 
                        apiBundle.getBundleContext();   
                bundleContextR.registerBundleListener(bundleListener);
            }
        } else 
        {
            BundleListener listener = new ConciergeBundleListener(this,bundleListener);
            super.bundleContext.addBundleListener(listener);
            this.bundleListeners.put(bundleListener, listener); 
        }
    }

    /**
     * {@inheritDoc}
     * 
     * cf. registerService method comment
     * 
     * @see org.objectweb.fractal.juliac.osgi.revision.BundleContextRevision#
     * registerServiceListener(org.objectweb.fractal.juliac.osgi.revision.event.ServiceListenerRevisionItf)
     */
    @Override
    public void registerServiceListener(ServiceListenerRevisionItf serviceListener)
    { 
        if(super.bundleContext.getBundle() == null)
        {
            BundleRevisionItf<Bundle,BundleContext> apiBundle = getBundle("bundle-frascati-api");
            if(apiBundle != null)
            {
                BundleContextRevisionItf<Bundle,BundleContext> bundleContextR = apiBundle.getBundleContext();   
                bundleContextR.registerServiceListener(serviceListener);
            }
        } else 
        {
            ServiceListener listener = new ServiceListenerRevision(serviceListener,this);
            super.bundleContext.addServiceListener(listener);
            this.serviceListeners.put(serviceListener, listener); 
        }
    }
    
    
    /**
     * {@inheritDoc}
     * 
     * cf. registerService method comment
     * 
     * @see org.objectweb.fractal.juliac.osgi.revision.api.BundleContextRevisionItf#
     * registerServiceListener(org.objectweb.fractal.juliac.osgi.revision.event.ServiceListenerRevisionItf,
     *  java.lang.String)
     */
    @Override
    public void registerServiceListener(ServiceListenerRevisionItf serviceListener, String filter)
    {
        if(super.bundleContext.getBundle() == null)
        {
            BundleRevisionItf<Bundle,BundleContext> apiBundle = getBundle("bundle-frascati-api");
            if(apiBundle != null)
            {
                BundleContextRevisionItf<Bundle,BundleContext> bundleContextR = apiBundle.getBundleContext();   
                bundleContextR.registerServiceListener(serviceListener);
            }
        } else 
        {
            ServiceListener listener = new ServiceListenerRevision(serviceListener,this);
            if(this.validateFilter(filter))
            {
                try
                {
                    super.bundleContext.addServiceListener(listener,filter);
                    this.serviceListeners.put(serviceListener, listener);
                    
                } catch (InvalidSyntaxException e)
                {
                    logger.log(Level.CONFIG,e.getMessage(),e); 
                }
            }
        }       
    }
    
    /**
     * {@inheritDoc}
     * 
     * cf. registerService method comment
     * 
     * @see org.objectweb.fractal.juliac.osgi.revision.api.BundleContextRevisionItf#
     * registerFrameworkListener(org.objectweb.fractal.juliac.osgi.revision.event.FrameworkListenerRevisionItf)
     */
    @Override
    public void registerFrameworkListener(
            FrameworkListenerRevisionItf frameworkListener)
    {
        if(super.bundleContext.getBundle() == null)
        {
            BundleRevisionItf<Bundle,BundleContext> apiBundle = getBundle("bundle-frascati-api");
            if(apiBundle != null)
            {
                BundleContextRevisionItf<Bundle,BundleContext> bundleContextR = apiBundle.getBundleContext();   
                bundleContextR.registerFrameworkListener(frameworkListener);
            }
        } else 
        {
            FrameworkListener listener = new FrameworkListenerRevision(frameworkListener);
            super.bundleContext.addFrameworkListener(listener);
            this.frameworkListeners.put(frameworkListener, listener);
        }        
    }
    
    /**
     * {@inheritDoc}
     * 
     * @see org.objectweb.fractal.juliac.osgi.revision.api.BundleContextRevisionItf#
     * mapBundleEvent(int)
     */
    @Override
    public String mapBundleEvent(int bundleEventType)
    {
        String bundleEventName = null;
        
        switch(bundleEventType)
        {
            case BundleEvent.INSTALLED: bundleEventName = "INSTALLED";
            break;
            case ConciergeBundleListener.CONCIERGE_RESOLVED_EVENT : bundleEventName = "RESOLVED";
            break;
            case BundleEvent.STARTED: bundleEventName = "STARTED";
            break;
            case BundleEvent.STOPPED: bundleEventName = "STOPPED";
            break;
            case BundleEvent.UNINSTALLED: bundleEventName = "UNINSTALLED";
            break;
            case BundleEvent.UPDATED : bundleEventName = "UPDATED";
            break;
            default : 
        }
        return bundleEventName;
    }

    /**
     * @param bundle
     * @return
     */
    protected BundleContext getBundleContext(Bundle bundle)
    {
        try
        {
            Field contextField = bundle.getClass().getDeclaredField("context");
            contextField.setAccessible(true);
            BundleContext context = (BundleContext) contextField.get(bundle);
            return context;
            
        } catch (SecurityException e)
        {
            logger.log(Level.CONFIG,e.getMessage(),e); 
            
        } catch (NoSuchFieldException e)
        {
            logger.log(Level.CONFIG,e.getMessage(),e); 
            
        } catch (IllegalArgumentException e)
        {
            logger.log(Level.CONFIG,e.getMessage(),e); 
            
        } catch (IllegalAccessException e)
        {
            logger.log(Level.CONFIG,e.getMessage(),e); 
        }
        return null; 
    }
    
    /**
     * @param string
     */
    private BundleRevisionItf<Bundle,BundleContext> getBundle(String symbolicName)
    {
        BundleRevisionItf<Bundle,BundleContext> searchedBundle =
            super.getRegisteredBundle(symbolicName);
        
        if(searchedBundle == null)
        {
            Bundle[] bundles = super.bundleContext.getBundles();
            for(Bundle bundle : bundles)
            {
                @SuppressWarnings("unchecked")
                Dictionary<String,String> headers = bundle.getHeaders();
                String bundleSymbolicName = headers.get("Bundle-SymbolicName");
                
                if(bundleSymbolicName != null && symbolicName.equals(bundleSymbolicName))
                {
                    BundleContext context = getBundleContext(bundle);
                                       
                    ConciergeBundleContext newContext = new ConciergeBundleContext(
                       (ConciergeRevision) super.getRevision(), context, getClassLoader());
                    
                    searchedBundle = new BundleRevision(getClassLoader(), newContext,bundle);
                    super.registerBundle((AbstractBundle<Bundle, BundleContext>) searchedBundle);
                    break;
                }            
            }
        }
        return searchedBundle;
    }
}
